#include "client.h"

void
do_syserr(char *file, int line)
{
  char buf[32];
  snprintf(buf, sizeof(buf), "error (%s line %d)", file, line);
  perror(buf);
  exit(1);
}

void
yourfault(char *str)
{
  fprintf(stderr, "%s\n", str);
  abort();
}

int
read_number(char *buf, int n, int i, int *result)
{
  while (i < n && (buf[i] == ' ' || buf[i] == 1))
    i++;
  if (i >= n) return -1;
  *result = 0;
  while (i < n && (buf[i] >= '0' && buf[i] <= '9')) {
    *result = *result * 10 + (buf[i] - '0');
    i++;
  }
  return i;
}

char *
read_configuration(int *port, int *password)
{
# define MAXBUF 80
  static char buf[MAXBUF];
  int fd, n, i;
  char *path = getenv("LISP_CONNECTION_FILE");

  if (!path) {
	char *rel = ".lep/default";
  	char *home;
	
	home = getenv("HOME");
	path = malloc(strlen(home)+strlen(rel)+2);
	if (!path) syserr();

	sprintf(path, "%s/%s", home, rel);
  }
  if ( (fd = open(path, O_RDONLY)) == -1) syserr();
  if ( (n = read(fd, buf, sizeof(buf))) == -1) syserr();
  if (n >= sizeof(buf)) yourfault("parse error");
  close(fd);
  buf[n] = 0;
  
  i = 0;
  if ( (i = read_number(buf, n, i, port)) == -1) yourfault("parse error");
  if ( (i = read_number(buf, n, i, password)) == -1) yourfault("parse error");

  return buf;
}

FILE *
do_connect(int port)
{
  int fd;
  struct sockaddr_in addr;
  addr.sin_family = AF_INET;
  addr.sin_port = htons(port);
  addr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
  if ( (fd = socket(PF_INET, SOCK_STREAM, 0)) == -1) syserr();
  if (connect(fd, (struct sockaddr *) &addr, sizeof(addr)) == -1) syserr();
  return fdopen(fd, "r+");
}
