#include "client.h"

#define MAX(a, b) ((a) > (b) ? (a) : (b))

typedef struct string {
  int max, len;
  char *data;
} string;

static void
append0(string *str, char *new, int n)
{
  if (str->max < str->len + n) {
    str->max += MAX(n, str->max);
    str->data = realloc(str->data, str->max);
  }
  memcpy(str->data + str->len, new, n);
  str->len += n;
}

static void
append1(string *str, char new)
{
  append0(str, &new, 1);
}

static void
append(string *str, char *new)
{
  append0(str, new, strlen(new));
}

static void
login(FILE *s, int password)
{
  fprintf(s, ":lep \"eval.c PID %d\" %d\n", getpid(), password);
  fprintf(s, "\"eval.c PID %d\"\n", getpid());
  fflush(s);
}

static void
login_stream(FILE *s, int password, char *sid)
{
  fprintf(s, ":stream \"eval.c PID %d\" %d %s %d\n",
	  getpid(), password, sid, -getpid());
  fflush(s);
}

#define STR_OMIT_QUOTES 1
#define STR_DOUBLE_ENCODE 2

static void
print_string(FILE *s, char *b, int n, int flags)
{
  int i;
  if (!(flags & STR_OMIT_QUOTES)) fputc('"', s);
  for (i = 0; i < n; i++) {
    char c = b[i];
    switch (c) {
    case '\\':
    case '"':
      if (flags & STR_DOUBLE_ENCODE) fputs("\\\\", s);
      fputc('\\', s);
      /* fallthrough */
    default:
      fputc(c, s);
    }
  }
  if (!(flags & STR_OMIT_QUOTES)) fputc('"', s);
}

static int
get_token(FILE *s, string *outbuf)
{
  int c;
  int string = 0;
  int quote = 0;

  outbuf->len = 0;
  if (feof(s)) return 0;

  /* skip whitespace */
  while ( (c = fgetc(s)) != EOF)
    if (!(c == ' '|| c == '\n'))
      break;
  if (c == EOF)
    return ferror(s) ? -1 : 0;
  ungetc(c, s);

  /* find whitespace or paren; special treatment for string and backslash. */
  while ( (c = fgetc(s)) != EOF)
    if (quote) {
      append1(outbuf, c);
      quote = 0;
    } else if (string)
      switch (c) {
      case '\\': quote = 1; break;
      case '"': goto done;
      default: append1(outbuf, c);
      }
    else
      switch (c) {
      case '\\': quote = 1; break;
      case '"': string = 1; break;
      case ' ': case '\t': case '\r': case '\n': goto done;
      case '(': case ')':
	if (outbuf->len == 0)
	  append1(outbuf, c);
	else
	  ungetc(c, s);
	goto done;
      default: append1(outbuf, c);
      }
done:
  if (c == EOF && ferror(s)) return -1;
  append1(outbuf, 0);
  return outbuf->len;
}

static void
send_eval_request(FILE *s, char *b, int n)
{
  fputs("(nil"
	" lep::make-session"
	" lep::evaluation-request"
	" :buffer-package :cl-user"
	" :session-id 0"
	" :text ",
	s);
  print_string(s, b, n, 0);
  fputs(")\n", s);
  fflush(s);
}

static void
usage(FILE *s, int rc)
{
  fputs("usage: eval [--rc] FORM\n"
	"  --rc    return integer result as exit code\n"
	"", s);
  exit(rc);
}

#define assert_equalp(got, expected) do_assert_equalp(__LINE__, got, expected)

static void
do_assert_equalp(int line, char *got, char *expected)
{
  if (!strcasecmp(got, expected)) return;
  fprintf(stderr,
	  "error (line %d): expected `%s', but got `%s'\n",
	  line, expected, got);
  exit(1);
}

int
main(int argc, char **argv)
{
  int port, password;
  FILE *s;
  int parse_rc = 0;
# define MAXBUF 80
  char buf[MAXBUF];
  string token = { 0 };
  int n;
  int neednewline = 0;
  
  if (argc >= 2 && !strcmp(argv[1], "--rc")) {
    argc--;
    argv++;
    parse_rc = 1;
  }
  if (argc != 2)
    usage(stderr, 1);
  if (!strcmp(argv[1], "--help"))
    usage(stdout, 0);

  read_configuration(&port, &password);
  s = do_connect(port);
  login(s, password);
  send_eval_request(s, argv[1], strlen(argv[1]));

  for (;;) {
#   define grab()							\
      do {								\
	if ( (n = get_token(s, &token)) < 0) syserr();			\
	else if (n == 0) yourfault("eof");				\
      } while(0)
    grab(); assert_equalp(token.data, "(");
    grab();
    if (!strcasecmp(token.data, "nil")) {
      char *id;
      FILE *bg;
      int nfds;
      struct pollfd fds[3];

      /* (nil :request lep::make-session-for-lisp 0 nil nil lep::create-listener-stream nil) */
      grab(); assert_equalp(token.data, ":request");
      grab(); assert_equalp(token.data, "lep::make-session-for-lisp");
      grab(); id = strdup(token.data);
      grab(); assert_equalp(token.data, "nil");
      grab(); assert_equalp(token.data, "nil");
      grab(); assert_equalp(token.data, "lep::create-listener-stream");
      grab(); assert_equalp(token.data, "nil");
      grab(); assert_equalp(token.data, ")");

      /* open :stream connection */
      bg = do_connect(port);
      login_stream(bg, password, id);

      /* shuffle data:
       *   - from standard input to stream
       *   - from stream to standard output
       *   - until data (i.e., a final reply) is ready on the main connection
       */
      fds[0].fd = fileno(s);
      fds[0].events = POLLIN;
      fds[1].fd = fileno(bg);
      fds[1].events = POLLIN;
      fds[2].fd = 0;
      fds[2].events = POLLIN;
      nfds = 3;
      do {
	fds[0].revents = fds[1].revents = fds[2].revents = 0;
	if (poll(fds, nfds, -1) < 0) syserr();
	if (fds[1].revents) {
	  if ( (n = read(fds[1].fd, buf, sizeof(buf))) < 0) syserr();
	  if (n == 0) break;
	  if (write(1, buf, n) != n) syserr();
	  neednewline = buf[n-1] != '\n';
	}
	if (fds[2].revents) {
	  if ( (n = read(0, buf, sizeof(buf))) < 0) syserr();
	  if (!n) {
	    nfds--;
	    shutdown(fds[1].fd, SHUT_WR);
	  } else
	    if (write(fds[1].fd, buf, n) != n) syserr();
	}
      } while (!fds[0].revents || fds[1].revents);
    } else {
      grab();
      if (!strcasecmp(token.data, ":error")) {
	/* (0 :error "condition") */
	grab();
	if (neednewline) putchar('\n');
	fprintf(stderr, "Error: %s\n", token.data);
	exit(2);
      }
      /* (0 :reply "result") */
      assert_equalp(token.data, ":reply");
      grab();
      if (parse_rc)
	return atoi(token.data);
      else { 
	if (token.data[n-1] == '\n') token.data[n-1] = 0; /* trim newline */
	if (neednewline) putchar('\n');
	puts(token.data);
      }
      break;
    }
  }
  return 0;
}
