(in-package :lep)

(defmacro with-gensyms ((&rest symbols) &body body)
  `(let (,@(loop
	       for symbol in symbols
	       collect `(,symbol (copy-symbol ',symbol))))
     ,@body))

(defmacro with-collector
    ((&optional (collect 'collect) &key result nconc append) &body body)
  (unless collect (setf collect '#:collect))
  (check-type collect symbol)
  (check-type result symbol)
  (with-gensyms (head tail)
    `(let* ((,head (cons nil '()))
	    (,tail ,head))
       (labels ((,collect (frob)
		  (setf ,tail (setf (cdr ,tail) (cons frob '())))
		  frob)
		,@(when nconc
		    `((,nconc (frobs)
			      (setf ,tail (last (setf (cdr ,tail) frobs))))))
		,@(when append
		    `((,append (frobs)
			       (dolist (frob frobs)
				 (,collect frob))))))
	 (symbol-macrolet (,@(when result
			       `((,result (cdr ,head)))))
	   ,@body))
       ,@(unless result
	   `((cdr ,head))))))

(defmacro with-reducer ((smasher function &optional (initial-value nil ivp))
			&body body)
  (check-type smasher symbol)
  (with-gensyms (tmp fn nada)
    `(let ((,tmp ,(if ivp initial-value `',nada))
	   (,fn ,function))
       (flet ((,smasher (frob)
		(if (eq ,tmp ',nada)
		    (setf ,tmp frob)
		    (setf ,tmp (funcall ,fn ,tmp frob)))))
	 ,@body)
       (if (eq ,tmp ',nada)
	   (funcall ,fn)
	   ,tmp))))

(defmacro for-finish ()
  '(loop-finish))

(defun for-aux (kind clauses body-form finally-forms)
  `(loop ,@(loop
	       for firstp = t then nil
	       for %clauses = clauses then (rest %clauses)
	       for clause = (first %clauses) then (first %clauses)
	       while (and %clauses (listp clause))
	       append (cons (ecase kind
			      (for (if firstp 'as 'and))
			      (for* 'as))
			    clause)
	       into result
	       finally (return (append result %clauses)))
       do (progn ,body-form)
       finally (progn) ,@finally-forms))

(defmacro %for ((&rest clauses) body-form &rest finally-forms)
  (for-aux 'for clauses body-form finally-forms))

(defmacro %for* ((&rest clauses) body-form &rest finally-forms)
  (for-aux 'for* clauses body-form finally-forms))

(defmacro for ((&rest clauses) &rest body-forms)
  `(%for ,clauses
       (progn ,@body-forms)))

(defmacro for* ((&rest clauses) &rest body-forms)
  `(%for* ,clauses
       (progn ,@body-forms)))

(defmacro while (condition &rest body-forms)
  `(for (:while ,condition)
     ,@body-forms))

(defmacro until (condition &rest body-forms)
  `(for (:until ,condition)
     ,@body-forms))

(defmacro do-sequence ((elt seq) &body body)
  `(block nil
     (map nil (lambda (x)
		,(if (listp elt)
		     `(destructuring-bind ,elt x ,@body)
		     `(let ((,elt x)) ,@body)))
	  ,seq)))

(defun listify (form)
  (if (listp form) form (list form)))

(defun string-suffix-p (str suffix)
  (and (<= (length suffix) (length str))
       (loop
	   for j :from 0 :below (length suffix)
	   and i :from (- (length str) (length suffix))
	   always (char= (char str i) (char suffix j)))))

(defmacro n++ (place &optional (delta 1) &environment env)
  "post-increment version of incf"
  (multiple-value-bind (dummies vals newval setter getter)
      (get-setf-expansion place env)
    `(let* (,@(mapcar #'list dummies vals) (,(car newval) ,getter))
       (prog1
	   ,(car newval)
	 (incf ,(car newval) ,delta)
	 ,setter))))

(defmacro aif (% then &optional else)
  `(let ((% ,%)) (if % ,then ,else)))

(defmacro awhen (% &body body)
  `(aif ,% (progn ,@body)))

(defun curry (fn &rest x)
  (lambda (&rest args)
    (apply fn (append x args))))

(defun rcurry (fn x)
  (lambda (&rest args)
    (apply fn (append args (list x)))))

(defun make-octets (n)
  (make-array n :element-type '(unsigned-byte 8)))
